# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
from torch.utils.data import Dataset, ConcatDataset
from datasets import load_dataset
from cosmos_rl.launcher.worker_entry import main as launch_worker
import cosmos_rl.utils.util as util
from cosmos_rl.policy.config import Config
from cosmos_rl.utils.util import basename_from_modelpath
from cosmos_rl.policy.config import Config as CosmosConfig
from transformers import AutoTokenizer
import argparse
import toml

FPS = 1
MAX_PIXELS = 81920


class CosmosSFTDataset(Dataset):
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        """
        Called by launcher after being mounted
        """
        self.config = config
        self.tokenizer = tokenizer

        if config.train.train_policy.dataset.split:
            if isinstance(config.train.train_policy.dataset.split, list):
                dataset_list = []
                for split_name in config.train.train_policy.dataset.split:
                    dataset_list.append(self.dataset[split_name])
                self.dataset = ConcatDataset(dataset_list)
            else:
                assert isinstance(config.train.train_policy.dataset.split, str)
                self.dataset = self.dataset[config.train.train_policy.dataset.split]

        # get multi-modal files paths
        cosmos_cache_dir = os.environ.get(
            "COSMOS_CACHE", os.path.join(os.path.expanduser("~"), ".cache/cosmos/")
        )
        video_clips_path = os.path.join(
            cosmos_cache_dir,
            "datasets",
            basename_from_modelpath(config.train.train_policy.dataset.name),
            config.train.train_policy.dataset.subset,
            "video_clips",
        )
        if not os.path.exists(video_clips_path):
            raise FileNotFoundError(
                f"Dataset directory {video_clips_path} does not exist. Please check the dataset path."
            )
        mm_files_paths = {}
        for root, dirs, files in os.walk(video_clips_path):
            for file in files:
                if file.endswith((".mp4", ".avi", ".mov")):  # Common video extensions
                    mm_files_paths[file] = os.path.join(root, file)
        self.mm_files_paths = mm_files_paths

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> tuple[str, str]:
        """
        Return a tuple of (prompt, reference answer)
        """
        payload = self.dataset[idx]
        conversations = copy.deepcopy(payload["conversations"])

        for conv in conversations:
            if conv["role"] == "user":
                assert isinstance(conv["content"], str), "User message must be string"
                # Rewrite to support image/video tokens
                content = [
                    {
                        "type": "video",
                        "video": self.mm_files_paths[payload["video"].split("/")[-1]],
                        "max_pixels": MAX_PIXELS,
                        "fps": FPS,
                    },
                    {
                        "type": "text",
                        "text": conv["content"],
                    },
                ]
                conv["content"] = content

        return conversations


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, required=True)
    args = parser.parse_known_args()[0]
    with open(args.config, "r") as f:
        config = toml.load(f)
    config = Config.from_dict(config)

    # Each worker needs to prepare the data independently
    util.prepare_cosmos_data(
        dataset=config.train.train_policy.dataset, fps=FPS, max_pixels=MAX_PIXELS
    )

    def get_dataset(config: CosmosConfig) -> Dataset:
        dataset = load_dataset(
            config.train.train_policy.dataset.name,
            config.train.train_policy.dataset.subset,
        )
        # Prepare video files
        return CosmosSFTDataset(dataset)

    # It is best practice to pass the dataset as a factory function
    # so that the dataset can be loaded on demand. (Not all workers need it)
    launch_worker(
        dataset=get_dataset,
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

try:
    from grouped_gemm import backend
except ImportError:
    print(
        "grouped_gemm is not available. To enable grouped_gemm, please run:"
        "pip install git+https://github.com/fanshiqing/grouped_gemm@v1.1.4"
    )
    backend = None


class FakeGroupMMBackwardCheck(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x, w, m_offsets, out_dtype):
        ctx.save_for_backward(x, w, m_offsets)
        ctx.out_dtype = out_dtype
        return torch._grouped_mm(x, w, m_offsets, out_dtype=out_dtype)

    @staticmethod
    def backward(ctx, grad_output):
        x, w, m_offsets = ctx.saved_tensors
        out_dtype = ctx.out_dtype

        grad_x = torch._grouped_mm(
            grad_output, w.transpose(-2, -1), m_offsets, out_dtype=out_dtype
        )
        grad_w = torch._grouped_mm(
            x.transpose(-2, -1), grad_output, m_offsets, out_dtype=out_dtype
        )
        # Fault due to grouped_gemm, where aligned memory is not used which may contains nan
        # Check https://github.com/pytorch/pytorch/issues/154557
        grad_x = torch.nan_to_num(grad_x, nan=0.0)
        grad_w = torch.nan_to_num(grad_w, nan=0.0)
        return grad_x, grad_w, None, None


class FallbackGroupedGemmImpl(torch.autograd.Function):
    @staticmethod
    def forward(ctx, a, b, batch_sizes, trans_b):
        assert backend is not None, "grouped_gemm is not available."
        assert (
            torch.count_nonzero(batch_sizes) != 0
        ), "Input batch_sizes should not be all zeros!"
        ctx.save_for_backward(a, b, batch_sizes)
        ctx.trans_b = trans_b
        return backend.gmm(a, b, batch_sizes, trans_a=False, trans_b=trans_b)

    @staticmethod
    def backward(ctx, grad):
        grad = grad.contiguous()
        a, b, batch_sizes = ctx.saved_tensors
        trans_b = ctx.trans_b

        agrad = None
        if ctx.needs_input_grad[0]:
            agrad = backend.gmm(
                grad, b, batch_sizes, trans_a=False, trans_b=not trans_b
            )

        bgrad = None
        if ctx.needs_input_grad[1]:
            lhs, rhs = (grad, a) if trans_b else (a, grad)
            bgrad = backend.gmm(lhs, rhs, batch_sizes, trans_a=True, trans_b=False)
        # Fault due to grouped_gemm, where aligned memory is not used which may contains nan
        agrad = torch.nan_to_num(agrad, nan=0.0)
        bgrad = torch.nan_to_num(bgrad, nan=0.0)
        return agrad, bgrad, None, None


def run_group_gemm_hopper(
    contig_tokens, m_sizes, m_offsets, gate_weight, up_weight, down_weight, act_fn
):
    gate_proj = FakeGroupMMBackwardCheck.apply(
        contig_tokens,
        gate_weight.transpose(-2, -1).contiguous(),
        m_offsets,
        torch.bfloat16,
    )

    up_proj = FakeGroupMMBackwardCheck.apply(
        contig_tokens,
        up_weight.transpose(-2, -1).contiguous(),
        m_offsets,
        torch.bfloat16,
    )

    # Apply activation
    hidden_outputs = act_fn(gate_proj) * up_proj

    # Run the third GEMM (down projection)
    hidden_outputs = FakeGroupMMBackwardCheck.apply(
        hidden_outputs,
        down_weight.transpose(-2, -1).contiguous(),
        m_offsets,
        torch.bfloat16,
    )
    return hidden_outputs


def run_group_gemm_3rd_party(
    contig_tokens, m_sizes, m_offsets, gate_weight, up_weight, down_weight, act_fn
):
    sizes_cpu = m_sizes.cpu().to(torch.int64)

    gate_proj = FallbackGroupedGemmImpl.apply(
        contig_tokens,
        gate_weight.transpose(-2, -1).contiguous(),
        sizes_cpu,
        False,
    )

    up_proj = FallbackGroupedGemmImpl.apply(
        contig_tokens,
        up_weight.transpose(-2, -1).contiguous(),
        sizes_cpu,
        False,
    )

    # Apply activation
    hidden_outputs = act_fn(gate_proj) * up_proj

    # Run the third GEMM (down projection)
    hidden_outputs = FallbackGroupedGemmImpl.apply(
        hidden_outputs,
        down_weight.transpose(-2, -1).contiguous(),
        sizes_cpu,
        False,
    )
    return hidden_outputs


use_torch_group_gemm_impl = None


def group_gemm_imp():
    global use_torch_group_gemm_impl
    if use_torch_group_gemm_impl is None:
        cuda_device_props = torch.cuda.get_device_properties()
        # currently torch._grouped_mm is only available on nightly build
        use_torch_group_gemm_impl = (
            cuda_device_props.major == 9
            and cuda_device_props.minor == 0
            and hasattr(torch, "_grouped_mm")
        )

    # Use torch implemetation if torch._grouped_mm is available, otherwise use 3rd party implementation
    return (
        run_group_gemm_hopper if use_torch_group_gemm_impl else run_group_gemm_3rd_party
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.policy.model.gpt.weight_converter import (
    convert_weight_from_hf as gpt_weight_from_hf,
)
from cosmos_rl.utils.parallelism import ParallelDims
import torch
from typing import Tuple


def map_key_from_hf(name: str, src_model_type: str) -> str:
    if src_model_type in ["llama", "qwen2", "qwen3", "qwen2_5_vl"]:
        return name.replace("model.", "")
    else:
        raise ValueError(f"Unsupported model type: {src_model_type}")


def convert_weight_from_hf(
    tensor: torch.Tensor,
    name: str,
    src_model_type: str,
    parallel_dims: ParallelDims,
    ignore_unknown_weights: bool = False,
) -> Tuple[str, torch.Tensor]:
    lm_part_name, lm_part_shard = gpt_weight_from_hf(
        tensor, name, src_model_type, parallel_dims, ignore_unknown_weights=True
    )
    if lm_part_name is not None:
        return lm_part_name, lm_part_shard
    assert name.startswith("visual."), f"Unsupported weight: {name}"

    if (
        parallel_dims.dp_shard_enabled
        or parallel_dims.cp_enabled
        or parallel_dims.tp_enabled
    ):
        dp_shard_rank = parallel_dims.mesh["dp_cp_tp"].get_local_rank()
        dp_shard_size = parallel_dims.mesh["dp_cp_tp"].size()
    else:
        dp_shard_rank = 0
        dp_shard_size = 1

    dest_name = name.replace("visual.", "")
    shard = tensor
    # TODO(cjx): Only FSDP sharding is supported for visual part
    shard = shard.contiguous()
    if shard.shape[0] % dp_shard_size == 0:
        shard = shard.tensor_split(dp_shard_size, dim=0)
        shard = shard[dp_shard_rank]
    else:
        chunk_size = (shard.shape[0] + dp_shard_size - 1) // dp_shard_size
        shard = shard[dp_shard_rank * chunk_size : (dp_shard_rank + 1) * chunk_size]

    return dest_name, shard.contiguous()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import triton
import triton.language as tl


__all__ = ["generate_permute_indices"]


# parallelized kernel
@triton.jit
def _fill_indices_kernel(
    tokens_per_expert_group_ptr,
    start_index_values_ptr,
    write_offsets_ptr,
    output_ptr,
    experts_per_rank: tl.constexpr,
    num_ranks: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,  # Number of threads per block
):
    pid = tl.program_id(axis=0)
    num_programs = tl.num_programs(axis=0)

    # map programs (blocks) to the experts and loop (grid stride) if needed
    for expert_id in range(pid, experts_per_rank, num_programs):
        # read this experts write offset
        write_offset = tl.load(write_offsets_ptr + expert_id)

        for r in range(num_ranks):
            # index into tokens_per_expert_group array
            i = r * experts_per_rank + expert_id

            # load start index and number of tokens for this expert-rank pair
            start_index = tl.load(start_index_values_ptr + i)
            length = tl.load(tokens_per_expert_group_ptr + i)

            # each thread in block processes tokens in parallel
            offsets = tl.arange(0, BLOCK_SIZE)

            # tokens are processed in chunks of BLOCK_SIZE
            for chunk_start in range(0, length, BLOCK_SIZE):
                chunk_offsets = chunk_start + offsets

                # mask valid indices
                mask = chunk_offsets < length

                values = start_index + chunk_offsets

                # destination
                dest_indices = write_offset + chunk_offsets

                # store
                tl.store(output_ptr + dest_indices, values, mask=mask)

            # update write offset for next rank
            write_offset += length


# ==============
# wrapper
# ==============


def fill_indices_wrapper(
    tokens_per_expert_group: torch.Tensor,
    start_index_values: torch.Tensor,
    write_offsets: torch.Tensor,
    experts_per_rank: int,
    num_ranks: int,
    total_size: int,
    block_size: int = 128,
    max_blocks: int = 1024,
):
    # Allocate exact size needed instead of max_len
    permuted_indices = torch.full(
        (total_size,), -1, dtype=torch.int32, device=tokens_per_expert_group.device
    )

    # write offsets is per local expert...
    num_blocks = min(experts_per_rank, max_blocks)
    # grid = one block per expert unless capped and then we loop...
    grid = (num_blocks,)

    # launch kernel
    _fill_indices_kernel[grid](
        tokens_per_expert_group,
        start_index_values,
        write_offsets,
        permuted_indices,
        experts_per_rank,
        num_ranks,
        BLOCK_SIZE=block_size,
    )
    return permuted_indices


# used for reference testing only


def fill_indices_cpu(
    tokens_per_expert_group: torch.Tensor,
    start_index_values: torch.Tensor,
    write_offsets: torch.Tensor,
    experts_per_rank: int,
    num_ranks: int,
    total_size: int,  # Changed from max_len to actual required size
):
    # Allocate exact size needed
    permuted_indices = torch.full(
        (total_size,),
        -1,
        dtype=torch.int32,
    )

    # Fill the permuted indices
    for e in range(experts_per_rank):
        write_start = write_offsets[e].item()
        for r in range(num_ranks):
            i = r * experts_per_rank + e
            start_index = start_index_values[i].item()
            length = tokens_per_expert_group[i].item()
            if length > 0:
                end_idx = min(write_start + length, total_size)
                permuted_indices[write_start:end_idx] = torch.arange(
                    start_index,
                    start_index + (end_idx - write_start),
                    dtype=torch.int32,
                )
            write_start += length
    return permuted_indices


def generate_permute_indices(
    tokens_per_expert_group: torch.Tensor,
    experts_per_rank: int,
    num_ranks: int,
    alignment: int,
    use_cpu: bool = False,
):
    """
    Prepare permutation indices and the number of tokens for each expert.
    Modified version that returns a tensor of size sum(m_sizes) instead of max_len.

    Args:
        tokens_per_expert_group: number of tokens for each expert from all ranks.
        experts_per_rank: number of experts per rank.
        num_ranks: number of ranks.
        alignment: alignment for each returned element in `m_sizes` and padding min for zero token experts.
        use_cpu: whether to use CPU implementation.

    Returns:
        permuted_indices: Tensor of indices with size sum(m_sizes), that map original token order to the expert-grouped order.
        m_sizes: aligned number of tokens for each expert (padded to alignment boundary).
        m_offsets: Cumulative sum of m_sizes. The exclusive ending position for each expert's tokens.

    Explanatory details:
        `tokens_per_expert_group` is of shape (num_ranks * experts_per_rank,), for example:
        From: |       rank 0      |       rank 1      |
        To:   | E0 | E1 | E2 | E3 | E0 | E1 | E2 | E3 |
              |  4 |  2 |  1 |  3 |  1 |  2 |  3 |  4 |
    """

    # prefix sum to get start index of each expert
    start_index_values = (
        torch.cumsum(tokens_per_expert_group, 0) - tokens_per_expert_group
    )

    # total tokens for each expert (sum over ranks)
    total_tokens_per_expert = tokens_per_expert_group.view(num_ranks, -1).sum(0)

    # pad out empty experts to alignment requirement
    total_tokens_per_expert = torch.clamp_min(total_tokens_per_expert, alignment)

    # align the chunk sizes (cdiv)
    m_sizes = ((total_tokens_per_expert + alignment - 1) // alignment * alignment).to(
        torch.int32
    )

    # additional prefix sum to get write offset of each expert in permuted_indices
    m_offsets = torch.cumsum(m_sizes, 0)
    write_offsets = m_offsets - m_sizes

    # Calculate the actual total size needed
    total_size = m_offsets[-1]

    # Select the implementation to use
    if use_cpu:
        permuted_indices = fill_indices_cpu(
            tokens_per_expert_group,
            start_index_values,
            write_offsets,
            experts_per_rank,
            num_ranks,
            total_size,
        )
    else:  # gpu
        permuted_indices = fill_indices_wrapper(
            tokens_per_expert_group,
            start_index_values,
            write_offsets,
            experts_per_rank,
            num_ranks,
            total_size,
        )

    return permuted_indices, m_sizes, m_offsets.to(torch.int32)


# Below is for testing only


def simple_test():
    device = torch.device("cuda", 0)
    experts_per_rank = 4
    num_ranks = 4
    tokens_per_expert_group = torch.full(
        (num_ranks * experts_per_rank,), 4, dtype=torch.int32, device=device
    )
    alignment = 32
    # Use the GPU kernel
    permuted_indices_gpu, m_sizes, _ = generate_permute_indices(
        tokens_per_expert_group,
        experts_per_rank,
        num_ranks,
        alignment,
        use_cpu=False,
    )
    # Use the CPU method
    permuted_indices_cpu, m_sizes, _ = generate_permute_indices(
        tokens_per_expert_group,
        experts_per_rank,
        num_ranks,
        alignment,
        use_cpu=True,
    )
    # Check that the results are the same

    assert torch.equal(permuted_indices_gpu.cpu(), permuted_indices_cpu)
    assert torch.equal(
        torch.remainder(m_sizes, alignment),
        torch.zeros(experts_per_rank, device=device),
    )
    # Print the results
    print(f"{permuted_indices_gpu=}, \n{permuted_indices_cpu=}")
    print(f"{m_sizes=}")
    print("Success")
    return True  # assert would have failed meaning getting here is success.


def test_with_zero_tokens():
    device = torch.device("cuda", 0)
    experts_per_rank = 4
    num_ranks = 2

    # Create a test case where some experts have zero tokens
    tokens_per_expert_group = torch.tensor(
        [4, 0, 2, 3, 1, 0, 0, 5],  # Some experts have zero tokens
        dtype=torch.int32,
        device=device,
    )

    alignment = 8

    # Use the GPU kernel
    permuted_indices_gpu, m_sizes, m_offsets = generate_permute_indices(
        tokens_per_expert_group,
        experts_per_rank,
        num_ranks,
        alignment,
        use_cpu=False,
    )

    # Use the CPU method
    permuted_indices_cpu, m_sizes_cpu, m_offsets_cpu = generate_permute_indices(
        tokens_per_expert_group,
        experts_per_rank,
        num_ranks,
        alignment,
        use_cpu=True,
    )

    # Check that the results are the same
    assert torch.equal(permuted_indices_gpu.cpu(), permuted_indices_cpu)
    assert torch.equal(m_sizes, m_sizes_cpu)

    # Verify that experts with zero tokens have at least min_slots_per_expert
    total_tokens_per_expert = tokens_per_expert_group.view(num_ranks, -1).sum(0)
    zero_token_experts = total_tokens_per_expert == 0
    if zero_token_experts.any():
        assert (m_sizes[zero_token_experts] >= alignment).all()

    # Check alignment
    assert torch.equal(
        torch.remainder(m_sizes, alignment),
        torch.zeros(experts_per_rank, device=device),
    )

    # Print the results
    print(f"tokens_per_expert_group = {tokens_per_expert_group}")
    print(f"total_tokens_per_expert = {total_tokens_per_expert}")
    print(f"m_sizes = {m_sizes}")
    print(f"m_offsets = {m_offsets}")
    print(f"permuted_indices = {permuted_indices_gpu[:sum(m_sizes).item()]}")

    # Check that experts with zero tokens have -1 in their slots
    for e in range(experts_per_rank):
        start = (m_offsets[e] - m_sizes[e]).item()
        end = m_offsets[e].item()
        expert_indices = permuted_indices_gpu[start:end]
        if total_tokens_per_expert[e] == 0:
            assert (
                expert_indices == -1
            ).all(), f"Expert {e} with zero tokens should have all -1 indices"
            assert (
                expert_indices.size(0) >= alignment
            ), f"Expert {e} with zero tokens should have at least {alignment} slots"
            print(
                f"Expert {e} has zero tokens and {expert_indices.size(0)} slots with all -1"
            )

    print("All tests passed successfully!")
    return True


if __name__ == "__main__":
    simple_test()
    test_with_zero_tokens()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.dispatcher.algo.base import RuleBasedAlgo, _register_rule_based_algo
from typing import Callable, List
import numpy as np
from cosmos_rl.utils.constant import Algo


class GRPO(RuleBasedAlgo):
    def __init__(
        self, reward_fn: Callable, unbiased: bool = False, eps: float = 1e-5, **kwargs
    ):
        super().__init__(**kwargs)
        self.reward_fn = reward_fn
        self.unbiased = unbiased
        self.eps = eps

    def compute_reward(self, to_be_evaluated: str, reference: str) -> float:
        return self.reward_fn.compute_reward(to_be_evaluated, reference)

    def compute_advantage(self, rewards: List[float]) -> List[float]:
        rewards = np.array(rewards).astype(np.float32)
        assert rewards.ndim == 1, "rewards must be a 1D array"
        mean = np.mean(rewards)
        if self.unbiased:
            result = rewards - mean
        else:
            result = (rewards - mean) / (np.std(rewards) + self.eps)

        return result.tolist()

    def ready(self) -> bool:
        return self.reward_fn is not None


_register_rule_based_algo(Algo.GRPO, GRPO)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.utils.parallelism import ParallelDims
import torch
import re
from typing import Tuple


def map_key_from_hf(name: str, src_model_type: str) -> str:
    if src_model_type in ["llama", "qwen2", "qwen2_5_vl", "qwen3"]:
        return name.replace("model.", "")
    else:
        raise ValueError(f"Unsupported model type: {src_model_type}")


def convert_weight_from_hf(
    tensor: torch.Tensor,
    name: str,
    src_model_type: str,
    parallel_dims: ParallelDims,
    ignore_unknown_weights: bool = False,
) -> Tuple[str, torch.Tensor]:
    tp_rank, tp_size = parallel_dims.tp_coord

    if parallel_dims.dp_shard_enabled or parallel_dims.cp_enabled:
        dp_shard_rank = parallel_dims.mesh[tuple(("dp_shard_cp",))].get_local_rank()
        dp_shard_size = parallel_dims.mesh[tuple(("dp_shard_cp",))].size()
    else:
        dp_shard_rank = 0
        dp_shard_size = 1

    dest_name = map_key_from_hf(name, src_model_type)

    if "lm_head.weight" == dest_name:
        shard = tensor.tensor_split(tp_size, dim=0)[tp_rank]
    elif "lm_head.bias" == dest_name:
        shard = tensor
    elif "embed_tokens.weight" == dest_name:
        shard = tensor.tensor_split(tp_size, dim=0)[tp_rank]
    elif dest_name in ["norm.weight", "norm.bias"]:
        shard = tensor
    elif (
        match := re.search(
            r"layers\.(\d+)\.self_attn\.(q_norm|k_norm|v_norm)\.(weight|bias)",
            dest_name,
        )
    ) is not None:
        shard = tensor
    elif (
        match := re.search(r"layers\.(\d+)\.input_layernorm\.(weight|bias)", dest_name)
    ) is not None:
        shard = tensor
    elif (
        match := re.search(
            r"layers\.(\d+)\.self_attn\.(q_proj|k_proj|v_proj)\.(weight|bias)",
            dest_name,
        )
    ) is not None:
        shard = tensor.tensor_split(tp_size, dim=0)[tp_rank]
    elif (
        match := re.search(
            r"layers\.(\d+)\.self_attn\.(o_proj)\.(weight|bias)", dest_name
        )
    ) is not None:
        if dest_name.endswith(".bias"):
            shard = tensor
        else:
            shard = tensor.tensor_split(tp_size, dim=-1)[tp_rank]
    elif (
        match := re.search(  # noqa: F841
            r"layers\.(\d+)\.mlp\.(up_proj|gate_proj)\.(weight|bias)", dest_name
        )
    ) is not None:
        shard = tensor.tensor_split(tp_size, dim=0)[tp_rank]
    elif (
        match := re.search(r"layers\.(\d+)\.mlp\.down_proj\.(weight|bias)", dest_name)  # noqa: F841
    ) is not None:
        if dest_name.endswith(".bias"):
            shard = tensor
        else:
            shard = tensor.tensor_split(tp_size, dim=-1)[tp_rank]
    elif (
        match := re.search(  # noqa: F841
            r"layers\.(\d+)\.post_attention_layernorm\.(weight|bias)", dest_name
        )
    ) is not None:
        shard = tensor
    elif not ignore_unknown_weights:
        raise ValueError(f"Unsupported weight: {dest_name}")
    else:
        return None, None

    # Do FSDP sharding
    shard = shard.contiguous()
    if shard.shape[0] % dp_shard_size == 0:
        shard = shard.tensor_split(dp_shard_size, dim=0)
        shard = shard[dp_shard_rank]
    else:
        chunk_size = (shard.shape[0] + dp_shard_size - 1) // dp_shard_size
        shard = shard[dp_shard_rank * chunk_size : (dp_shard_rank + 1) * chunk_size]

    return dest_name, shard.contiguous()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from cosmos_rl.utils.logging import logger
import os
import shutil


def modelscope_download_model(name: str, **kwargs):
    """
    Download a model from ModelScope.
    Args:
        name (str): The name of the model to download.
        **kwargs: Additional arguments to pass to the snapshot_download function.
    """
    cache_dir = os.environ.get(
        "MODELSCOPE_CACHE",
        os.path.join(os.path.expanduser("~"), ".cache/modelscope_cache/model"),
    )
    local_dir = os.path.join(cache_dir, name)
    try:
        from modelscope.hub.snapshot_download import snapshot_download

        if int(os.environ.get("LOCAL_RANK", 0)) == 0:
            if not os.path.exists(local_dir):
                snapshot_download(
                    name,
                    local_dir=local_dir,
                    repo_type="model",
                    max_workers=16,
                    **kwargs,
                )
    except Exception as e:
        logger.error(f"Failed to download model {name} from ModelScope: {e}")
        # Remove all in foler local_dir
        if int(os.environ.get("LOCAL_RANK", 0)) == 0:
            if os.path.exists(local_dir):
                shutil.rmtree(local_dir)

    return local_dir


def modelscope_download_dataset(dataset_name: str, **kwargs):
    """
    Download a dataset from ModelScope.
    Args:
        name (str): The name of the dataset to download.
        **kwargs: Additional arguments to pass to the snapshot_download function.
    """
    cache_dir = os.environ.get(
        "MODELSCOPE_CACHE",
        os.path.join(os.path.expanduser("~"), ".cache/modelscope_cache/dataset"),
    )
    local_dir = os.path.join(cache_dir, dataset_name)
    try:
        from modelscope.hub.snapshot_download import snapshot_download

        if int(os.environ.get("LOCAL_RANK", 0)) == 0:
            if not os.path.exists(local_dir):
                allow_patterns = None
                if "subset_name" in kwargs:
                    subset_name = kwargs.pop("subset_name")
                    allow_patterns = [f"{subset_name}/*", f"{subset_name}/**/*", "*"]
                snapshot_download(
                    dataset_name,
                    local_dir=local_dir,
                    repo_type="dataset",
                    max_workers=16,
                    allow_patterns=allow_patterns,
                    **kwargs,
                )
    except Exception as e:
        logger.error(f"Failed to download dataset {dataset_name} from ModelScope: {e}")
        # Remove all in foler local_dir
        if int(os.environ.get("LOCAL_RANK", 0)) == 0:
            if os.path.exists(local_dir):
                shutil.rmtree(local_dir)
    if int(os.environ.get("LOCAL_RANK", 0)) == 0:
        if os.path.exists(os.path.join(local_dir, "dataset_infos.json")):
            os.remove(os.path.join(local_dir, "dataset_infos.json"))

    return local_dir


def modelscope_download(name: str, repo_type: str = "model", **kwargs):
    """
    Download a model or dataset from ModelScope.

    Args:
        name (str): The name of the model or dataset to download.
        repo_type (str): The type of repository, either 'model' or 'dataset'.
        **kwargs: Additional arguments to pass to the snapshot_download function.
    """
    if os.environ.get("COSMOS_USE_MODELSCOPE", False) in [True, "True", "true", "1"]:
        if repo_type == "model":
            return modelscope_download_model(name, **kwargs)
        elif repo_type == "dataset":
            return modelscope_download_dataset(name, **kwargs)
    return name


def update_config_if_modelscope(loaded_config: Any):
    """
    Update the model and dataset paths in the loaded configuration to point to the modelscope storage.
    This is necessary for compatibility with the modelscope storage system.
    """
    loaded_config.policy.model_name_or_path = modelscope_download(
        loaded_config.policy.model_name_or_path, "model"
    )
    loaded_config.train.train_policy.dataset.name = modelscope_download(
        loaded_config.train.train_policy.dataset.name,
        "dataset",
        subset_name=loaded_config.train.train_policy.dataset.subset,
        revision=loaded_config.train.train_policy.dataset.revision,
    )

    if loaded_config.validation.dataset.name:
        loaded_config.validation.dataset.name = modelscope_download(
            loaded_config.validation.dataset.name,
            "dataset",
            subset_name=loaded_config.validation.dataset.subset,
            revision=loaded_config.validation.dataset.revision,
        )

    return loaded_config


def modelscope_load_dataset(dataset_name: str, subset_name: str, split: str):
    """
    Load dataset from ModelScope.
    """
    if os.environ.get("COSMOS_USE_MODELSCOPE", False) in [True, "True", "true", "1"]:
        from modelscope.msdatasets import MsDataset

        if os.path.exists(dataset_name):
            from datasets import load_dataset

            return load_dataset(dataset_name, subset_name, split=split)
        return MsDataset.load(dataset_name, subset_name=subset_name, split=split)
    return None
